#!/usr/bin/python
# -*- mode: python; coding: utf-8 -*-

# atheist
#
# Copyright (C) 2009 David Villa Alises
#
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA




import os, sys, string
import glob
import subprocess as subp
import optparse
import logging
import time
import select
import signal
import thread, threading
import copy
import traceback
import tempfile
import types
from itertools import chain

"""
Debugging level:

0:
[  OK  ]( 0) samples/test1.test              ls
[ FAIL ]( 2) samples/test1.test              ls na
Total: 0.03s - 1/2

1: INFO
samples/test1.test: 2 tests
[  OK  ]( 0) samples/test1.test              ls
[ FAIL ]( 2) samples/test1.test              ls na
Total: 0.54s - 1/2

2: DEBUG
samples/test1.test loading
samples/test1.test: 2 tests
Executing <Test 'test1.test' - 'ls'>
<Test 'test1.test' - 'ls'> process ends
Executing <Test 'test1.test' - 'ls na'>
<Test 'test1.test' - 'ls na'> process ends
[  OK  ]( 0) samples/test1.test              ls
[ FAIL ]( 2) samples/test1.test              ls na
Total: 0.03s - 1/2
"""

VERSION = '0.'+'$Date: 2009-09-06 00:30:52 +0200 (Sun, 06 Sep 2009) $'[7:17].replace('-','')
USAGE = '''%%prog [options] [file|directory]

%%prog is a testing tool.
This is version %s, Copyright (C) 2009 David Villa Alises
%%prog comes with ABSOLUTELY NO WARRANTY; This is free software, and you are
welcome to redistribute it under certain conditions; See COPYING for details.''' % VERSION


class MyFormatter(logging.Formatter):
    def format(self, record):
        record.levelinitial = record.levelname[0]*2
        return logging.Formatter.format(self, record)


log = logging.getLogger('atheist')
log.setLevel(logging.WARNING)

GREY = chr(27)+'[38m'
RED =   chr(27)+'[31m'
GREEN = chr(27)+'[32m'

IGREEN = chr(27)+'[7;32m'
NORM =  chr(27)+'[m'

BOLD = chr(27)+'[1m'

OK =   GREEN + "OK" + NORM
FAIL = RED + "FAIL" + NORM
NONE = GREY + "??" + NORM

RESULT_STR = {None: NONE, True: OK, False: FAIL}

SETUP    = '_setup.test'
TEARDOWN = '_teardown.test'

# FIXME: fichero de configuración .atheist
EXCLUDE_DIRS = ['.svn', '.hg']
EXCLUDE_FILES = [SETUP, TEARDOWN]

abort = False


def beaty_value(value):
    return RESULT_STR[value]

def ellipsis(cad, width=30, just=False):
    if not cad: cad = ''
    retval = cad.split('\n')[0][:width-1]
    retval = unicode(retval, 'utf-8')
    #if retval != cad: retval += "…"
    if retval != cad: retval += "_"
    if just: retval = retval.ljust(width)
    return retval


def compath(path):
    return path.replace(config.base_dir, '.')

def count(n, unit):
    return "%s %s%s" % (n, unit, 's' if n>1 else '')

def unique_filename(prefix=''):
    return "/tmp/atheist_%s_%s_%s" % \
           (prefix, os.getpid(), str(time.time()).replace('.',''))


def file_template():
    return '''# -*- mode:python; coding:utf-8 -*-

# Test(cmd,
#      cwd         = '.',
#      delay       = 0,
#      desc        = 'test template',
#      detach      = False,
#      env         = {},
#      expected    = 0,
#      must_fail   = False,
#      path        = '.',
#      template    = [],
#      timeout     = 5,
#      save_stdout = False,
#      signal      = signal.SIGKILL)
'''


class Record(object):
    "kargs are automatic attributes"

    def __init__(self, **kargs):
        self.__dict__.update(kargs)
        self._attrs = kargs.keys()

    def __getattr__(self, attr):
        if not self.__dict__.has_key(attr):
            raise AttributeError, attr
        return self.__dict__.get(attr, None)

    def __setattr__(self, attr, value):
        self.__dict__[attr] = value

    def __str__(self):
        retval = ""
        for k in self._attrs:
            retval += "%s:%s " % (k, self.__dict__[k])

        return "<Record %s>" % retval


class ThreadFunc(threading.Thread):
    def __init__(self, function, args=(), kargs={}, start=True):
        threading.Thread.__init__(self, name=function.__name__)
        self.function = function
        self.args = args
        self.kargs = kargs
        self.active = threading.Event()
        self.active.set()
        if start:
            self.start()

    def cancel(self):
        print '*' * 20
        log.debug("ThreadFunc.cancel '%s'" % self.function.__name__)
        self.active.clear()

    def run(self):
        self.function(*self.args, **self.kargs)


class ProgressBar:
    def __init__(self, max_val=100, label=''):
        self.max = max_val       # total item amount
        self.label = label       # prefix for the bar
        self.width = 42          # bar width
        self.step = max_val/100
        self.val = 0             # current item amount
        self.blk = 0
        self.pct = 0             # completed percentage

        if config.verbosity or config.stdout or max_val<4:
            self._render = lambda x:None

        self.tinit = time.time()
        self.eta = ''

    def set_width(self, value):
        self.width = value - 7

    def inc(self, val=1, cad=''):
        self.val += val
        self.blk = self.val * self.width / self.max
        self.pct = self.val * 100 / self.max
        if self.val > 3:
            self.eta = "%2.1fs" % (((self.max  - self.val) * (time.time() - self.tinit)) / self.val)
        self._render(cad)

    def _render(self, ustr):
        cad = ':%4s [ %s' % (self.label[:4], self.blk * '#')
        cad += (self.width - self.blk) * '-'
        cad += ' ] %s/%s (%3d%%) %s %s\r' % (str(self.val).rjust(len(str(self.max))),
                                              self.max, self.pct, self.eta,
                                              ellipsis(ustr))
        print cad,
        sys.stdout.flush()

    def clean(self):
        print (' ' * (self.width + 24 + len(self.label))) + '\r',


def print_exc(exc):
    for line in exc.split('\n'):
        if line.strip(): log.error('| ' + line)


class Stats(Record):
    def __init__(self):
        self.tini = time.time()
        Record.__init__(self, total=0, ok=0)

    def __str__(self):
        t = "%.2fs" % (time.time() - self.tini)
        if self.ok == self.total:
            return (BOLD + IGREEN + " ALL OK!! " + NORM + " - %s - %s " ) % (t, count(self.total, 'test'))

        return "%s - %s/%s" % (t, self.ok, count(self.total, 'test'))



class TypedList(list):
    def __init__(self, cls):
        self._cls = cls

    def append(self, val):
        log.error("Use operator += instead")

    def __iadd__(self, val):
        if isinstance(val, list):
            for i in val: self += i
            return self

        assert isinstance(val, self._cls), \
               ("arg must be a '%s'" % self._cls.__name__)
        list.append(self, val)
        return self


class Condition:
    def __init__(self, cads=[]):
        self._value = None
        self._cads  = cads

    def run(self):
        raise NotImplemented

    def evaluate(self):
        if self._value == None:
            self._value = self.run()
        return self._value

    def expand_vars(self, vars):
        for c in self._cads:
            setattr(self, c, string.Template(getattr(self,c)).safe_substitute(vars).strip())

    def clone(self):
        retval = copy.copy(self)
        retval._value = None
        return retval

    def name(self):
        return self.__class__.__name__

    def __repr__(self):
        return str(self)

    def __str__(self):
        return "<%s[%s] %s>" % \
            (self.name(), beaty_value(self._value), self.basic_info())

    def summary(self):
        return "%s %s>" % (str(self)[:-1], self.more_info())

    def basic_info(self):
        return ''

    def more_info(self):
        return ''


class ConditionList(list):
    def append(self, val):
        log.error("Use operator += instead")
        return

    def __iadd__(self, cond):
        assert isinstance(cond, Condition), "arg must be a Condition"
        list.append(self, cond)
        return self


class EnvVarDefined(Condition):
    def __init__(self, varname, value=None):
        Condition.__init__(self)
        self.varname = varname
        self.value = value

    def run(self):
        retval = self.varname in os.environ.keys()
        if retval and self.value:
            retval = os.environ[self.varname] == self.value
        return retval

    def basic_info(self):
        return "'%s'" % self.varname


class FileExists(Condition):
    def __init__(self, fname):
        self.fname = fname
        Condition.__init__(self)

    def run(self):
        return os.path.exists(self.fname)

    def basic_info(self):
        return "'%s'" % self.fname


class DirExists(FileExists):
    def run(self):
        return FileExists.run(self) and os.path.isdir(self.fname)


class FileContains(Condition):
    def __init__(self, fname, contains):
        self.fname = fname
        self.contains = contains
        Condition.__init__(self)

    def run(self):
        if not os.path.exists(self.fname):
            log.info("'%s' does not exists" % compath(self.fname))
            return False

        fd = file(self.fname)
        fcontent = fd.read()
        fd.close()
        return self.contains in fcontent

    def basic_info(self):
        return "'%s' content:'%s'" % (self.fname, self.contains[:20])


class OpenPort(Condition):
    def __init__(self, port, host='localhost', proto='tcp'):
        assert proto in ['tcp','udp']
        assert isinstance(port , int) and 0 < port < 65535
        Condition.__init__(self)

        self.proto = proto
        self.port = port
        self.host = host

    def run(self):
        if self.host == 'localhost':
            self.cmd = 'fuser -n %s %s > /dev/null 2> /dev/null' % \
                       (self.proto, self.port)
        else:
            self.cmd = 'nmap -p%s %s | grep "%s/%s open" > /dev/null' % \
                       (self.port, self.host, self.port, self.proto)
        return os.system(self.cmd) == 0

    def basic_info(self):
        return "'%s/%s'" % (self.proto, self.port)

    def more_info(self):
        return "cmd: %s" % self.cmd


class TaskRunning(Condition):
    def __init__(self, task):
        assert isinstance(task, Task)
        self.task = task
        Condition.__init__(self)

    def run(self):
        return self.task.ps and self.task.ps.poll() is None

    def basic_info(self):
        return "'%s'" % self.task.name


class ProcessRunning(Condition):
    def __init__(self, pid):
        assert isinstance(pid, int)
        self.pid = pid
        Condition.__init__(self)

    def run(self):
        return os.waitpid(self.pid, os.WNOHANG) != (0,0)

    def basic_info(self):
        return "'%s'" % self.pid


class DebPkgInstalled(Condition):
    def __init__(self, package):
        assert isinstance(package, str)
        self.package = package
        Condition.__init__(self)

    def run(self):
        return os.system('dpkg -l %s 2> /dev/null| grep ^ii > /dev/null' % \
                             self.package) == 0

    def basic_info(self):
        return "'%s'" % self.package



class ConditionDecorator(Condition):
    def __init__(self, condition):
        assert isinstance(condition, Condition)
        self.condition = condition
        Condition.__init__(self)

    def expand_vars(self, vars):
        Condition.expand_vars(self.condition, vars)

    def name(self):
        return "%s:%s" % (self.__class__.__name__, self.condition.name())

    def basic_info(self):
        return self.condition.basic_info()

    def more_info(self):
        return self.condition.more_info()


class Not(ConditionDecorator):
    def run(self):
        return not self.condition.run()


class Poll(ConditionDecorator):
    def __init__(self, condition, interval=1, timeout=5):
        ConditionDecorator.__init__(self, condition)
        self.interval = interval
        self.timeout = timeout

    def run(self):
        enlapsed = 0
        while 1:
            log.debug("Polling condition: %s" % self.condition)
            if self.condition.run(): return True
            if enlapsed > self.timeout: break
            time.sleep(self.interval)
            enlapsed += self.interval

        return False


class RemoteTestFactory:
    def __init__(self, account):
        self.account = account
        try:
            self.user, self.host = self.account.split('@')
        except ValueError:
            self.host = self.account
            self.user = os.getlogin()

    def __call__(self, cmd, **values):
        values.update({'shell':True})
        return Test('ssh %s "%s"' % (self.account, cmd), **values)



DFL_TIMEOUT = 5
DFL_SIGNAL = signal.SIGKILL


task_attrs = {
    'cwd':         Record(type=str,  default=None),
    'delay':       Record(type=int,  default=0),
    'desc':        Record(type=str,  default=''),
    'detach':      Record(type=bool, default=False),
    'env':         Record(type=dict, default=os.environ.copy()),
    'expected':    Record(type=int,  default=0),
    'id':          Record(type=str,  default=None),
    'must_fail':   Record(type=bool, default=False),
    'path':        Record(type=str,  default=''),
    'template':    Record(type=list, default=[]),
    'timeout':     Record(type=int,  default=DFL_TIMEOUT),
    'save_stdout': Record(type=bool, default=False),
    'shell':       Record(type=bool, default=False),
    'signal':      Record(type=int,  default=DFL_SIGNAL),
    'stdout':      Record(type=str,  default=''),
    }


class Task(Record):

    @classmethod
    def validate_kargs(cls, _dict):
        for k,v in _dict.items():
            assert task_attrs.has_key(k), \
                "'%s' is not a valid keyword" % k
            assert isinstance(v, task_attrs[k].type), \
                "'%s' arg type must be %s" % (k, task_attrs[k].type)
            if v == task_attrs[k].default:
                log.warning("Default value for '%s' is already '%s'. Remove it." % (k, v))
        return _dict


    def __init__(self, cmd, **kargs):
        current_ids = [x.id for x in TC] + [x.id for x in ts]
        assert 'id' not in kargs.keys() or \
            kargs['id'] not in current_ids, "Duplicate task id: '%s'" % kargs['id']

        ts.append(self)
        self.cmd = cmd if cmd is not None else ''

        defaults = {}
        for k,v in task_attrs.items(): defaults[k] = v.default
        self.__dict__.update(defaults)

        if kargs.has_key('template'):
            for i in kargs['template']:
                self.__dict__.update(Task.validate_kargs(i))

        self.__dict__.update(Task.validate_kargs(kargs))

        #add the user variables to dft environ
        environ = os.environ.copy()
        for k,v in self.env.items():
            if environ.has_key(k) and v not in environ[k]:
                environ[k] += self.env[k]
            else:
                environ[k] = self.env[k]
        self.env = environ

        self.index = None
        self.fname = ''
        self.gen = TypedList(str)
        self.pre =  TypedList(Condition)
        self.post = TypedList(Condition)
        self.terminates = TypedList(Task)

        assert isinstance(self.cmd, str) or self.cmd is None

#        if not self.desc and self.cmd:
#            self.desc = self.cmd

        self.mode = None
        self.rawcmd = None
        self.result = None
        self.ps = None
        self.thread = None

        self.stdout_fd = None

        if self.stdout:
            self.stdout_fd = open(self.stdout, 'w')

        if self.save_stdout and not self.stdout:
            fd, self.stdout = tempfile.mkstemp(suffix='.stdout', prefix='atheist_')
            self.stdout_fd = os.fdopen(fd, 'w')

        if self.stdout:
            self.save_stdout = True
            self.post += FileExists(self.stdout)

        self.index = len(TC) + len(ts) - 1
        self.log = logging.getLogger("%s" % (self.name))
        term = logging.StreamHandler()
        term.setFormatter(MyFormatter(
                '[%(levelinitial)s] %(name)s: %(message)s',
                '%d/%m/%y %H:%M:%S'))
        term.setLevel(config.loglevel)
        self.log.setLevel(config.loglevel)
        self.log.addHandler(term)


    def kill(self, n):
        try:
            os.killpg(self.ps.pid, n)
        except OSError, e:
            log.debug("%s" % e)
        except AttributeError:
            self.log.warning("did not even start")
            return

        while 1:
            if self.ps.poll() is not None: return
            time.sleep(0.5)


    @property
    def description(self):
        return  "%s:%s" % (self.name, self.desc)

    def details(self):
        dic = self.__dict__.copy()
        dic['fname'] = compath(dic['fname'])

        keys = ['cmd']
#        if self.cmd != self.rawcmd: keys.append('rawcmd')
        keys.append('fname')
        if self.result is not None: keys.append('result')

        # add standard attrs with no-default value
        for k,v in task_attrs.items():
            if getattr(self,k) != v.default:
                keys.append(k)

        attrs = []
        for i in keys:
            attrs.append("%s: '%s'" % (i.rjust(14), dic[i]))

        # multivalued attributes
        for a in ['pre','post','gen','terminates']:
            for i in getattr(self, a):
                attrs.append("%s %s" % ('%s:' % a.rjust(14), i))

        return '%3s:%s\n%s\n' % (self.index, self.__class__.__name__,
                                str.join('\n', attrs))

    def __repr__(self):
        return str(self)

    def __str__(self):
        return "%s: <%s '%s' - '%s'>" % \
               (self.name, self.__class__.__name__,
                os.path.basename(self.fname), self.cmd.split('\n')[0])

    @property
    def name(self):
        return self.initial + str(self.index)

    @property
    def initial(self):
        return self.__class__.__name__[0]

    def value(self):
        return "[%s]" % beaty_value(self.result).center(len(FAIL))

    def result_summary(self):
        retcode = self.ps.returncode if self.ps else '-'
        return "  %s(%2s) %s%s%-3d %s | %s" % \
            (self.value(), retcode, self.mode,
             self.initial, self.index,
             ellipsis(self.cmd, 30, True),
             self.desc if self.desc != self.cmd else '')


    def expand_vars(self):
        vars = {
            'basedir': compath(config.base_dir),
            'dirname': compath(os.path.dirname(self.fname))}

        def substitute(val):
             return string.Template(val).safe_substitute(vars).strip()

        if self.cmd:
            self.rawcmd = self.cmd
            self.cmd = substitute(self.cmd)

        for p in self.pre + self.post:
            p.expand_vars(vars)

        for i in range(len(self.gen)):
            self.gen[i] = substitute(self.gen[i])


    def clean(self):
        "Remove generated files (by explicit declaration)"
        remove = [x for x in self.gen if os.path.exists(x)]
        if not remove: return

        self.log.info("removing generated files")
        for f in remove:
            if os.path.isfile(f):
                self.log.debug("- removing file '%s'" % compath(f))
                os.remove(f)
            elif os.path.isdir(f):
                self.log.warning("- removing directoy '%s'" % compath(f))
                if not f.startswith(config.base_dir):
                    self.log.error("cleto-mode disabled. Foreign directories will not be removed")
                    continue

                os.system('rm -r %s' % f)
            else:
                self.log.warning("%s is not a file nor directory!" % compath(f))


    def exec_cmd(self):

        def stdout_tag_write(text):
            for line in text.split('\n'):
                if not line: continue
                sys.stdout.write("%s| %s\n" % (self.name, line))
                sys.stdout.flush()

        def file_write(text):
            self.stdout_fd.write(text)
            self.stdout_fd.flush()

        out_funcs = []

        if config.stdout:
            out_funcs.append(stdout_tag_write)
        if self.stdout:
            out_funcs.append(file_write)

        if self.path:
            os.environ['PATH'] += ':'+self.path

        self.tini = time.time()

        if self.shell:
#            cmd = ('/bin/bash -c \"%s\"' % self.cmd.replace('\"', '\\"')).split()
            cmd = ['/bin/bash', '-c', "%s" % self.cmd]
        else:
            cmd = self.cmd.split()

#        print cmd
#        cmd = 'setsid /bin/sh -c \"%s\"' % self.cmd.replace('\"', '\\"')
#        cmd = self.cmd
        try:
            self.ps = subp.Popen(cmd,
                                 close_fds  = True,
                                 stdin      = subp.PIPE,
                                 stdout     = subp.PIPE,
                                 stderr     = subp.STDOUT,
                                 shell      = False,
                                 bufsize    = 0,
                                 cwd        = self.cwd,
                                 env        = self.env,
                                 preexec_fn = os.setsid)
        except OSError, e:
            self.log.error("%s: '%s'" % (e, self.cmd))
            self.result = False
            return

        self.log.info("starts (pid: %s)" % self.ps.pid)

        tini = time.time()
        read_fds = [self.ps.stdout]
        read_ready = select.select(read_fds,[],[], 0.05)[0]

        while self.ps.poll() is None or read_fds:
            for fd in read_ready:
                # http://mail.python.org/pipermail/python-bugs-list/2001-March/004491.html
                data = os.read(fd.fileno(), 2048)
                if not data:
                    read_fds.remove(fd)
                else:
                    for func in out_funcs: func(data)

            read_ready = select.select(read_fds,[],[], 0.5)[0]

            if self.timeout is None: continue

            if time.time()-tini > self.timeout:
                self.log.debug("timeout expired (%s>%s), sending signal %s to %s" % \
                                   (time.time() - self.tini, self.timeout,
                                    self.signal, self.ps.pid))
                self.kill(self.signal)


        self.log.debug("finish with %s" % self.ps.returncode)

        self.result = (self.ps.returncode == self.expected)
        if self.must_fail:
            self.result = self.ps.returncode != 0
    #/exec_cmd


class Test(Task): pass

class TestFunc(Test):
    def __init__(self, func, args, **kargs):
        assert callable(func)
        self.func = func
        self.args = args
        Test.__init__(self, func.__name__, **kargs)

    def exec_cmd(self):
        try:
            self.result = self.func(*self.args)
        except Exception,e:
            self.log.error(e)
            self.result = False

class Command(Task): pass

class Daemon(Command):
    def __init__(self, cmd, **kargs):

        forbidden = ['timeout','detach','expected']
        matches = set(forbidden) & set(kargs.keys())
        if matches:
            log.critical("You can not use '%s' with %s" % \
            (str.join(' or ', matches), self.__class__.__name__))
            sys.exit(1)

        Command.__init__(self, cmd, **kargs)
        self.timeout = None
        self.detach = True
        self.expected = -9


class Template(dict):
    def __init__(self, **kargs):
        self.update(Task.validate_kargs(kargs))




def run_test(test):
    if test.detach:
        test.log.debug("detaching")
        test.detach = False
        test.thread = ThreadFunc(run_test, (test,))
        time.sleep(0.2)
        return None

    time.sleep(test.delay)

    # pre-conditions
    for c in test.pre:
        value = c.evaluate()
        test.log.info("Pre-cond: %s" % c)
        if not value:
            test.result = False
            return False

    if test.cmd:
        test.exec_cmd()
    else:
        test.result = True

    if test.stdout_fd:
        test.stdout_fd.close()

    # post-conditions
    for c in test.post:
        value = c.evaluate()
        test.log.info("Post-cond: %s" % c)
        if not value:
            test.result = False
            return False

    # terminates
    for t in test.terminates:
        test.log.debug("terminates T%s sending signal %s" % \
                           (t.index, t.signal))
        Task.kill(t, t.signal)

    return test.result


class TaskCollection:

    def __init__(self):
        self.sets = []

    def __len__(self):
        return sum([len(s.tasks) for s in self.sets])

    def append_set(self, fname, tasks):
        self.sets.append(Record(fname=fname, tasks=tasks[:]))

    def __iter__(self):
        "Iterate over all tests"
        for t in chain(*[x.tasks for x in self.sets]):
            yield t


#class TaskSet:
#
#    def __init__(self):
#        self.files = []
#
#    def append_test(self, fname, task):
#        if not self.files or self.files[-1].fname != fname:
#            self.current_file = Record(fname=fname, tasks=[])
#            self.files.append(self.current_file)
#
#        self.current_file.tasks.append(task)
#
#    def __len__(self):
#        return sum([len(f.tasks) for f in self.files])
#
#    def __iter__(self):
#        for t in chain(*[f.tasks for f in self.files]):
#            yield t

TC = TaskCollection()

#n = 0
#TC = [] # todos
ts = [] # cargados en cada fichero

def get_task(_id):
    for t in chain(TC,ts):
        if t.id == _id: return t
    log.error("There is no task with id '%s'" % _id)
    return None


def process_script(code, testset, template={}, fname=''):

    before = testset[:]

    template.update({
        'basedir': compath(config.base_dir)})

    env = exec_env.copy()
    env['ts'] = testset

    code = string.Template(code).safe_substitute(template)

    try:
        exec(code) in env
    except Exception, e:
        log.error("Errors in the test definition '%s'" % fname)
        print_exc(traceback.format_exc())
        sys.exit(1)
        return

    for t in [x for x in testset if x not in before]:
        t.fname = fname



def load_file(fname):
    lines = file(fname).readlines()
    for line in lines:
        if line.startswith('include'):
            #print '-', line
            index = lines.index(line)
            #print index
            lines.remove(line)
            #print fname
            ilines = file(line.strip()[9:-2]).readlines()
            lines[index:index] = ilines

    return  str.join('', lines)


class task_mode:
    MAIN,SETUP,TEARDOWN = ['-', 's', 't']


def process_file(fname, mode=task_mode.MAIN):
    if config.list_only:
        print '-', compath(fname)
        return

    if mode == task_mode.MAIN:
        for i in ts[:]: ts.remove(i)  # empty the list :-S

    testdir =  os.path.dirname(fname)
    setup =    os.path.join(testdir, SETUP)
    teardown = os.path.join(testdir, TEARDOWN)

    if config.no_setup == False and \
            os.path.exists(setup) and \
            fname not in [setup, teardown]:
        process_file(setup, task_mode.SETUP)

    log.debug("%s loading" % compath(fname))

    aux_fname = os.path.splitext(compath(fname))[0]
    process_script(load_file(fname),
                   ts,
                   {'dirname':  compath(testdir),
                    'fname':    aux_fname,
                    'testname': os.path.basename(aux_fname)},
                   compath(fname)
                   )

    for t in ts:
        if t.mode is None: t.mode = mode

    log.info("%s: %s" % (compath(fname), count(len(ts), 'task')))

    if not ts:
        log.info("%s: does not contains any test" % compath(fname))
        return

    if config.no_setup == False and \
            os.path.exists(teardown) and \
            fname not in [setup, teardown]:
        process_file(teardown, task_mode.TEARDOWN)

    if mode == task_mode.MAIN:
        for t in ts:
            for i in t.gen:
                t.pre  += Not(FileExists(i))
                t.post += FileExists(i)

        TC.append_set(fname, ts)



def process_directory(dname):
    for root, dirs, files in os.walk(dname):
        log.info("Entering directory '%s'" % compath(root))
        for f in sorted([x for x in files if x.endswith(config.ext)]):
            if f in EXCLUDE_FILES: continue
            process_file(os.path.join(root, f))
        for d in EXCLUDE_DIRS:
            if d in dirs: dirs.remove(d)



def main():

    if config.inline:
        print config.inline
        fdn, fname = tempfile.mkstemp(suffix='.test', prefix='atheist_')
        fd = os.fdopen(fdn, 'w')
        fd.write(config.inline)
        fd.close()
        process_file(fname)


    for i in sorted([os.path.abspath(x) for x in args]):
        if os.path.isdir(i):
            process_directory(i)

        elif os.path.isfile(i):
            if i in [SETUP, TEARDOWN]: continue
            process_file(i)

        else:
            log.warning("Could not read file '%s'" % i)
            return 1


    if config.dry_run:
        for t in TC: print t.details()
        return


    if config.clean_only:
        for t in TC:
            t.clean()
        return


    pb = ProgressBar(len(TC), 'Run ')
    pb.set_width(int(os.environ.get('COLUMNS', 50)))
    for f in TC.sets:
        for t in f.tasks:
            if abort: break

            log.info(t)
            if run_test(t) == False and isinstance(t, Test) \
                    and not config.keep_going:
                t.log.info("FAIL, skipping remaining tasks (keep-going mode disabled)")
                break
            #pb.inc(1, t.description)
            pb.inc()

        for t in f.tasks:
            if config.clean: t.clean()

        while 1:
            unfinished = [x for x in f.tasks if x.thread \
                              and x.thread.isAlive() \
                              and x.timeout is not None]

            if not unfinished: break

            for t in unfinished:
                t.log.debug("waiting to finish detached")

                if abort:
                    Task.kill(t, signal.SIGKILL)

                t.thread.join(1)

        daemons = [x for x in f.tasks if x.thread and x.thread.isAlive() \
                       and x.timeout is None]
        if daemons:
            log.debug("-- Killing unterminated daemons...")
            for t in daemons:
                log.debug("sending signal %s to %s" % (t.signal, t.name))
                Task.kill(t, t.signal)

                t.thread.join(1)

    pb.clean()

    for taskset in TC.sets:
        print "Test case '%s'" % compath(taskset.fname)
        for task in taskset.tasks:
            print task.result_summary()
#            print "%s  -  %s" % (ellipsis(task.cmd), task.desc)



    tests = [x for x in TC if isinstance(x, Test)]
    stats.total += len(tests)
    stats.ok += len([x for x in tests if x.result])

    if stats.total:
        print "Total:", stats

    if stats.ok != stats.total:
        return 1

    return 0


def abort_handler(sign, frame):
    global abort
    print
    log.warning("C-c pressed!")
    abort = True


exec_env = {}
for c in [Test, TestFunc, Template,
          Command, Daemon,
          EnvVarDefined,
          FileExists, FileContains, DirExists,
          OpenPort,
          TaskRunning,
          DebPkgInstalled,
          Not, Poll,
          RemoteTestFactory,
          get_task]:
    exec_env[c.__name__] = c



if __name__ == '__main__':
    signal.signal(signal.SIGPIPE, signal.SIG_DFL)
    signal.signal(signal.SIGINT, abort_handler)
    signal.signal(signal.SIGTERM, abort_handler)

    parser = optparse.OptionParser(usage=USAGE)
    parser.add_option('-b', '--base-dir', dest='base_dir', default=os.getcwd(),
                      help='change working directory')
    parser.add_option('-c', '--clean', dest='clean', action='store_true',
                      help='remove generated files after test execution')
    parser.add_option('-C', '--clean-only', dest='clean_only',
                      action='store_true',
                      help="don't execute nothing, only remove generated files")

    parser.add_option('-e', '--stderr', dest='stderr', action='store_true',
                      help='print test process stderr')

    parser.add_option('-g', '--gen-template', dest='gen_template', action='store_true',
                      help='generate a test file template with default values')

    parser.add_option('-j', '--skip-setup', dest='no_setup', action='store_true',
                      help='skip _setup and _teardown files')

    parser.add_option('-k', '--keep-going', dest='keep_going',
                      action='store_true', default=False,
                      help='continue even with failed tests')

    parser.add_option('-l', '--list', dest='list_only', action='store_true',
                      help='list tests but do not execute them')
    parser.add_option('-n', '--dry-run', dest='dry_run', action='store_true',
                      help='print the commands that would be executed, but does not execute them')
    parser.add_option('-o', '--stdout', dest='stdout', action='store_true',
                      help='print test process stdout')

    parser.add_option('-s', '--script', dest='inline',
                      help='specifies command line script')

    parser.add_option('-t', '--time-tag', dest='timetag', action='store_true',
                      help='include time info in the logs')


    parser.add_option('-v', '--verbose', dest='verbosity',
                      default=0, action='count',
                      help='incresse verbosity')

    parser.add_option('-x', '--extension', dest='ext', default='.test',
                      help='extension for the test files')

    (config, args) = parser.parse_args()

    if config.gen_template:
        print file_template()
        sys.exit(0)

    if not args and not config.inline:
        parser.print_help()
        sys.exit(1)


    stats = Stats()

    config.loglevel = logging.WARNING
    if config.verbosity == 1:
        config.loglevel = logging.INFO
    elif config.verbosity >= 2:
        config.loglevel = logging.DEBUG

    log.setLevel(config.loglevel)


    if config.timetag:
        log_format = '[%(asctime)s] [%(levelinitial)s] %(message)s'
    else:
        log_format = '[%(levelinitial)s] %(message)s'

    console = logging.StreamHandler()
    console.setFormatter(MyFormatter(log_format, '%d/%m/%y %H:%M:%S'))
    log.addHandler(console)

    sys.exit(main())



